!========================================================
!--------------------------------------------------------
! Includes a module for the non-inclusive correction
! -------------------------------------------------------
!========================================================
module emsn_tools
  use types; use consts_dp
  use rad_tools; use hoppet_v1
  implicit none
  private

  public :: Mqq, Mgg
  public :: non_incl, dot

  real(dp), parameter :: lntwo =&
       & 0.693147180559945309417232121458176568075_dp

contains

  ! contains the analytic expressions for the non-inclusive (R-dependent)
  ! corrections
  function non_incl(R,part) result(res)
    real(dp), intent(in) :: R
    real(dp) :: correl, indep, res
    character(len=*) :: part

    ! Taylor series of the correlated emission contribution neglecting terms of O(R**8)
    ! Good agreement with MC below R=2.6
    !
    ! Taken from 1203.5773 A.16 (and independently derived by PFM)
    correl = -(-131._dp+12._dp*pisq+132._dp*lntwo)/72._dp*ca_def*log(R/1.74_dp) &
         & -(23.0_dp-24.0_dp*lntwo)/72._dp*nf*log(R/0.84_dp) &
         & +((1429._dp+3600._dp*pisq+12480._dp*lntwo)*ca_def+&
         & (3071._dp-1680._dp*lntwo)*nf)/172800._dp*R**2           &
         & +((-9383279._dp-117600._dp*pisq+1972320._dp*lntwo)*ca_def+&
         & two*(178080._dp*lntwo-168401._dp)*nf)/406425600._dp*R**4 &
         &  +((74801417._dp-33384960._dp*lntwo)*ca_def+&
         & (7001023._dp-5322240._dp*lntwo)*nf)/97542144000._dp*R**6

    if (R>pi) then
       ! Taken from 1206.4998 Eq.(45)
       indep = -half*A(1)*((pi/6._dp*R**2-1/pi/8._dp*R**4)*&
            & atan(pi/sqrt(R**2-pisq)) &!this avoids Floating point exceptions for R<pi
            & +(R**2/8._dp-pisq/12._dp)*sqrt(R**2-pisq))
    else
       ! Taken from 1203.5773 A.18 (and independently derived by PFM)
       indep = -half*A(1)*(pisq*R**2/12._dp-R**4/16._dp)
    end if

    select case(trim(part))
    case('correl')
       res = correl
    case('indep')
       res = indep
    case('all')
       res = correl+indep
    end select
  end function non_incl

  ! Below are reported the squared amplitudes for the double-soft current
  !-----------------------------------------------------------------------
  !-----------------------------------------------------------------------
  ! Eqs. (2.4) and (2.5a) divided by the double-independent eikonal amplitude squared
  function Mgg(p,pbar,k1,k2) result(res)
    real(dp), intent(in) :: p(:),pbar(:),k1(:),k2(:)
    real(dp) :: res
    !---------------------------------------------------
    real(dp) :: a1,b1,a2, b2, m2, atot, btot, shat
    real(dp) :: J2, W, invW1, invW2, S, R
    res = zero

    shat = dot(p,pbar)
    a1 = dot(p,k1);    a2=dot(p,k2)
    b1 = dot(pbar,k1); b2=dot(pbar,k2)
    atot = a1+a2; btot=b1+b2
    m2 = dot(k1,k2)

    if (a2 == zero .or. b2 == zero) then
       res = zero
       return
    end if

    J2 = two*((a1*b2-a2*b1)/m2/atot/btot)**2


    W = shat/atot/btot

    S = one/(a1*b2*m2)+one/(a2*b1*m2)-shat/(a1*b1*a2*b2)
    S = S*shat

    R = (a1*b2+a2*b1)/(atot*btot)

    invW1 = (a1*b1)/shat
    invW2 = (a2*b2)/shat

    res = S*(two-R)+J2-four*W/m2

    res = res*invW1*invW2

  end function Mgg

  ! Eqs. (2.4) and (2.5b) divided by the double-independent eikonal amplitude squared
  function Mqq(p,pbar,k1,k2) result(res)
    real(dp), intent(in) :: p(:),pbar(:),k1(:),k2(:)
    real(dp) :: res
    !---------------------------------------------------
    real(dp) :: a1,b1,a2, b2, m2, atot, btot,shat
    real(dp) :: J2, W, invW1, invW2
    res = zero

    shat = dot(p,pbar)
    a1 = dot(p,k1);    a2=dot(p,k2)
    b1 = dot(pbar,k1); b2=dot(pbar,k2)
    atot = a1+a2; btot=b1+b2
    m2 = dot(k1,k2)

    J2 = two*((a1*b2-a2*b1)/m2/atot/btot)**2

    W = shat/atot/btot

    invW1 = (a1*b1)/shat
    invW2 = (a2*b2)/shat


    res = -J2+two*W/m2

    res = res*invW1*invW2

  end function Mqq

  function dot(p,q) result(res)
    real(dp), intent(in) :: p(:),q(:)
    real(dp) :: res
    res = p(4)*q(4)-dot_product(p(:3),q(:3))
  end function dot

end module emsn_tools

