#!/usr/bin/env python

# {{{ imports
import os
import shutil
import glob
import math
import re
import sys
import copy
from os.path import join as pjoin
from collections import defaultdict
# }}}

# {{{ def: readin_file(file_path)


def readin_file(file_path):
    # this reads a file and writes it into a table
    with open(file_path, 'r') as f:
        table = [row.strip().split() for row in f if row.strip() and not row.split()[
            0].startswith("#") and not row.split()[0].startswith("%")]
    for row_index, row in enumerate(table):
        for column_index, column in enumerate(table[row_index]):
            table[row_index][column_index] = float(
                table[row_index][column_index])
    return table
# }}}
# {{{ def: add_table(table1,table2,norm=1)


def add_table(table1, table2, norm=1):
    # this function adds two tables
    norm = float(norm)

    # if the first table is actually no table
    if len(table1) == 0:
        table = normalize_table(table2, 1/norm)
        return table, False

    # fail safe if very first table is corrupted
    if len(table1) < len(table2) or len(table1[0]) < len(table2[0]):
        table = normalize_table(table2, 1/norm)
        return table, True

    # adding the tables
    table = [[0 for col in range(len(table1[0]))]
             for row in range(len(table1))]
    for row_index, row in enumerate(table1):
        for column_index, column in enumerate(table1[row_index]):
            if (column_index == 0 or column_index == 1):
                table[row_index][column_index] = table1[row_index][column_index]
            elif (column_index % 2 == 0):  # central values
                table[row_index][column_index] = table1[row_index][column_index] + \
                    norm*table2[row_index][column_index]
            else:  # errors
                table[row_index][column_index] = math.sqrt(
                    table1[row_index][column_index]**2 + norm**2*table2[row_index][column_index]**2)
    return table, False
# }}}
# {{{ def: normalize_table(table1,norm)


def normalize_table(table1, norm):
    # this function divides a table by a number
    norm = float(norm)
    table = [[0 for col in range(len(table1[0]))]
             for row in range(len(table1))]
    for row_index, row in enumerate(table1):
        for column_index, column in enumerate(table1[row_index]):
            if (column_index == 0 or column_index == 1):
                table[row_index][column_index] = table1[row_index][column_index]
            elif (column_index % 2 == 0):  # central values
                table[row_index][column_index] = table1[row_index][column_index]/norm
            else:  # errors
                table[row_index][column_index] = table1[row_index][column_index]/norm
    return table
# }}}
# {{{ def: get_number_of_events(run,scale,channel)


def get_number_of_events(run, scale, channel):
    # this reads a file and writes it into a table
    result_file_path = pjoin(run, "result", "result_"+channel+".dat")
    with open(result_file_path, 'r') as f:
        number_of_events = float(f.readline())
    return number_of_events
# }}}
# {{{ def: write_table_to_file(table,file_path)


def write_table_to_file(table, file_path, headline):
    # this writes a table to a file
    with open(file_path, 'w') as f:
        length = 30
        header = ""
        for item in headline:
            header = header+" "*(length-len(item))+item
        f.write(header+"\n")
        for row_index, row in enumerate(table):
            line = ""
            for column_index, column in enumerate(table[row_index]):
                line = line+" " * \
                    (length-len("%.16g" % table[row_index][column_index])
                     )+"%.16g" % table[row_index][column_index]
            f.write(line+"\n")
        f.close()


# }}}
# {{{ def: multidim_dict(n)
def multidim_dict(n):
    # extremely flexible multidimensional dictionary
    """ Creates an n-dimension dictionary where the n-th dimension is of type 'type'
    """
    return defaultdict(lambda: multidim_dict(n-1))
# }}}
# {{{ def: RadISH_run_combination(common_folder_name, final_output_folder = "", path = "", include_pre = "1")


def RadISH_run_combination(common_folder_name, final_output_folder="", path="", include_pre="1"):
    # combine RadISH output, callable from external codes
    if path:
        os.chdir(path)
    run_folders = sorted(glob.glob(common_folder_name+"*"))
    if not include_pre == "1":
        run_folders_tmp = copy.copy(run_folders)
        run_folders = []
        for folder in run_folders_tmp:
            if os.path.exists(pjoin(folder, "main.ident")):
                run_folders.append(folder)

    scales = sorted([x.rsplit("/", 1)[1]
                     for x in glob.glob(pjoin(run_folders[0], "RadISH_output", "scale*"))])
    histograms = sorted([x.rsplit("/", 1)[1] for x in glob.glob(
        pjoin(run_folders[0], "RadISH_output", scales[0], "*.dat"))])

    all_histograms_with_channels = multidim_dict(3)
    all_histograms_averaged_with_separate_channels = multidim_dict(2)
    all_histograms_combined = multidim_dict(2)
    number_of_events = multidim_dict(4)
    total_number_of_events = multidim_dict(3)
    previous_number_of_events = multidim_dict(3)

    # read in files
    for folder in run_folders:
        radish_folder = pjoin(folder, "RadISH_output")
        run = folder
        for scale in scales:
            scale_folder = pjoin(radish_folder, scale)
            for histogram in histograms:
                histogram_path = pjoin(scale_folder, histogram)
                if os.path.exists(histogram_path):
                    all_histograms_with_channels[os.path.basename(histogram_path)][os.path.basename(
                        os.path.dirname(histogram_path))][os.path.dirname(histogram_path)] = readin_file(histogram_path)
                    histogram_name = histogram.split("__")[0]
                    channel = histogram.split("__")[1].replace(
                        ".dat", "").replace("--lum", "")
                    if not channel in number_of_events[histogram_name][run][scale]:
                        number_of_events[histogram_name][run][scale][channel] = get_number_of_events(
                            run, scale, channel)
                        if channel in total_number_of_events[histogram_name][scale]:
                            total_number_of_events[histogram_name][scale][
                                channel] += number_of_events[histogram_name][run][scale][channel]
                        else:
                            total_number_of_events[histogram_name][scale][
                                channel] = number_of_events[histogram_name][run][scale][channel]

    # for scale in total_number_of_events:
    #     for channel in total_number_of_events[scale]:
    #         print "total_number_of_events["+scale+"]["+channel+"] = ", total_number_of_events[scale][channel]

    # average
    for histogram in all_histograms_with_channels:
        histogram_name = histogram.split("__")[0]
        channel = histogram.split("__")[1].replace(
            ".dat", "").replace("--lum", "")
        for scale in all_histograms_with_channels[histogram]:
            #            number_of_runs = len(all_histograms_with_channels[histogram][scale])
            for path in all_histograms_with_channels[histogram][scale]:
                new_path = pjoin(path.split("/", 1)[1])
                run = path.split("/")[0]
                try:
                    all_histograms_averaged_with_separate_channels[new_path][histogram], replaced_first_table = add_table(
                        all_histograms_averaged_with_separate_channels[new_path][histogram], all_histograms_with_channels[histogram][scale][path], number_of_events[histogram_name][run][scale][channel])
                    if replaced_first_table:
                        total_number_of_events[histogram_name][scale][
                            channel] -= previous_number_of_events[histogram_name][scale][channel]
                except:
                    total_number_of_events[histogram_name][scale][channel] -= number_of_events[histogram_name][run][scale][channel]
                previous_number_of_events[histogram_name][scale][channel] = number_of_events[histogram_name][run][scale][channel]
            # # divide by number of runs
        for new_path in all_histograms_averaged_with_separate_channels:
            scale = os.path.basename(new_path)
            all_histograms_averaged_with_separate_channels[new_path][histogram] = normalize_table(
                all_histograms_averaged_with_separate_channels[new_path][histogram], total_number_of_events[histogram_name][scale][channel])

    # add channels
    for path in all_histograms_averaged_with_separate_channels:
        for channel in all_histograms_averaged_with_separate_channels[path]:
            scale = path.split("/")[1]
            if "--lum" in channel:
                histogram = channel.split("__")[0]+"--lum.dat"
            else:
                histogram = channel.split("__")[0]+".dat"
            all_histograms_combined[scale][histogram], dummy = add_table(
                all_histograms_combined[scale][histogram], all_histograms_averaged_with_separate_channels[path][channel])

    # write out combined histograms
    if not final_output_folder:
        final_output_folder = "RadISH_combined"
    for scale in all_histograms_combined:
        try:
            os.makedirs(pjoin(final_output_folder, scale))
        except:
            pass
        for histogram in all_histograms_combined[scale]:
            out_file = pjoin(final_output_folder, scale, histogram)
            if "--lumi" in out_file:
                headline = ["dummy_low", "dummy_up",
                            "born0", "lumi1", "lumi2", "lumi3"]
            else:
                headline = [histogram.replace(".dat", "")+"_low", histogram.replace(
                    ".dat", "")+"_up", "res", "res_err", "exp1", "exp1_err", "exp2", "exp2_err", "exp3", "exp3_err"]
            write_table_to_file(
                all_histograms_combined[scale][histogram], out_file, headline)
# }}}


if __name__ == "__main__":
    common_folder_name = sys.argv[1]
    if len(sys.argv) > 2:
        final_output_folder = sys.argv[2]
    else:
        final_output_folder = ""
    RadISH_run_combination(common_folder_name, final_output_folder)
